/* $Workfile: LineMon.c $       $Revision: 2 $      $Modtime: 10/24/12 12:55p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    LineMon.c

    A simple example console application demonstrating the use of OTX Driver
    with a Thor/Sleipnir/Gimle board to monitor the E1 or T1 line for framing errors.

    The following Thor boards are supported by this application:

    Thor-2-PCMCIA (PRO and EX)
    Thor-2-PCI
    Thor-8-PCI
    Thor-2-PCI-Plus
    Thor-2-PCI-Express
    Thor-4-PCI-Plus
    Thor-8-PCI-Plus
    Gimle-16-PCI-Plus
    Sleipnir-1-PCI-Plus

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures 2 E1 Lines.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <assert.h>
#include <time.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxGeneric.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"
#include "OtxT1E1.h"
#include "OtxData.h"
#include "OtxDsp.h"

#define DEMO_DEFAULT_LOG_FILE "LineMon.log"
#define DEMO_DEFAULT_BOARD_NO 0


typedef struct DemoEnvT {
  char                  *pszInputStr;
  OTX_BOOL               bExitOnError;
  char                   cwd[OTX_MAX_PATH];
  char                   logFileName[OTX_MAX_PATH];
  OTX_INT32              nBoardType;
  int                    boardNo;
  int                    coLi;
  OtxT1E1LiModeE         eT1E1Mode;    // T1 or E1 mode
  OtxDataFormatE         eLaw;
  OTX_BOOL               bReceiveEqualizer;
  OtxT1LineCodeE         eT1LineCode;
  OtxT1FrameFormatE      eT1FrameFormat;
  OtxE1LineCodeE         eE1LineCode;
  OtxE1FrameFormatE      eE1FrameFormat;
  OtxBrdClockSourceE     eClkSource;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(const char *formatStr, ...);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);

static OTX_INT32 DemoProcessUserCommand(char commandChar);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;

struct PhysDevs {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hQuadE1T1[4]; 
  OTX_HANDLE hLi[16];
  OTX_HANDLE hCodec[4];
} g_hPhysDevs;


struct T1E1CfgData {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;


/*******************************************************************************

  @func

  Toggles the T1E1 Xmit Pin Control bit

*******************************************************************************/
void DemoToggleT1E1XmitPinControl(void)
{
  static OTX_UINT32 bTxPinEnable[8] = { 0 };
  OTX_INT32 i;
  OTX_RESULT result;

  for (i = 0; i < g_demoEnv.coLi; i++) {
    bTxPinEnable[i] ^= 1;
    result = OtxT1E1XmitPinControl(g_hPhysDevs.hLi[i], (OTX_BOOL) bTxPinEnable[i]); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    DemoLogf("OtxT1E1XmitPinControl for Line #%d set %s\n", i, bTxPinEnable[i] ? "TRUE" : "FALSE");
  }
}


/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;
                               
  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = (OTX_EVENT) NULL;   
    result = OTX_S_OK;
  }

  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(g_hEventQueue, result);
}



/*******************************************************************************

  @func
    Open the Physical Devices on the board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1;
  OTX_UINT32 nDeviceId;
  OTX_UINT32 nSubsystemId;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCMCIA:
    DemoLogf("Opening Thor-2-PCMCIA board\n");
    break;
  case OTX_DEVICE_THOR_PCI:
    DemoLogf("Opening Thor-2-PCI board\n");
    break;
  case OTX_DEVICE_THOR_2_PCI_PLUS:
    DemoLogf("Opening Thor-2-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    DemoLogf("Opening Thor-2-PCI-Express board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI:
    DemoLogf("Opening Thor-8-PCI board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    DemoLogf("Opening Thor-8-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    DemoLogf("Opening Thor-4-PCI-Plus board\n");
    break;
  case OTX_DEVICE_SLEIPNIR_1_PCI:
    DemoLogf("Opening Sleipnir-1-PCI board\n");
    break;
  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
    DemoLogf("Opening Gimle-16-PCI-Plus board\n");
    break;
  case OTX_DEVICE_GIMLE_16_PCIE:
    DemoLogf("Opening Gimle-16-PCIe board\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_DEVICE_ID, &nDeviceId); DemoCheckRc(g_hPhysDevs.hThorController0, result);
  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SUBSYSTEM_ID, &nSubsystemId); DemoCheckRc(g_hPhysDevs.hThorController0, result);
  DemoLogf("Board Device ID: 0x%.4X, SubsystemId: 0x%.4X\n", nDeviceId, nSubsystemId);

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCMCIA:
  case OTX_DEVICE_THOR_PCI:
  case OTX_DEVICE_SLEIPNIR_1_PCI:
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    // On the Thor-2-PCI and Thor-2-PCMCIA boards the E1T1 devices are hosted by the board itself. Get handles to T1/E1 line interface devices #0 and #1
    for (i = 0; i < g_demoEnv.coLi; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }
    break;

  case OTX_DEVICE_THOR_8_PCI:
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    // On the Thor-8-PCI board the E1T1 devices are hosted by two quad E1T1 device. Open those two devices.
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }

    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      nQuadE1T1 = i/4;
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
    break;
  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
  case OTX_DEVICE_GIMLE_16_PCIE:
    // On the Gimle-16-PCI-Plus board the E1T1 devices are hosted by four quad E1T1 devices. Open those 4 devices.
    for (nQuadE1T1 = 0; nQuadE1T1 < 4; nQuadE1T1++) {
      DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }

    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      nQuadE1T1 = i/4;
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    // On the Thor-4-PCI board the E1T1 devices are hosted by a quad E1T1 device. Open those two devices.
    for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
      DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }

    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      nQuadE1T1 = i/4;
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }

  
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCI:
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
  case OTX_DEVICE_THOR_8_PCI:
  case OTX_DEVICE_THOR_4_PCI_PLUS:
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    // Get handles to Codec devices #0,#1,#2 and #3
    for (i = 0; i < 4; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_CODEC, i, i, g_hEventQueue, &g_hPhysDevs.hCodec[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);

      // Set PCM compression law for the Codec
      DemoLogf("Set Law to %s\n", g_demoEnv.eLaw == OTX_DATA_FORMAT_COMPAND8_A_LAW ? "A-law" : "U-law");
      result = OtxBrdCodecSetLaw(g_hPhysDevs.hCodec[i], g_demoEnv.eLaw); DemoCheckRc(g_hPhysDevs.hCodec[i], result);

    }
    break;
  case OTX_DEVICE_THOR_PCMCIA:
  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
  case OTX_DEVICE_GIMLE_16_PCIE:
  case OTX_DEVICE_SLEIPNIR_1_PCI:
    // No Codec on these boards
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }

}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  // Configure Transceivers
  for (i = 0; i < g_demoEnv.coLi; i++) {

    
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    } else {
      // E1 mode
      result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }
    result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(g_hPhysDevs.hLi[i], result);

    // Set termination for the T1/E1 Front End
    result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_HIGHZ); DemoCheckRc(g_hPhysDevs.hLi[i], result);

    // Tri-state the transmit pins of the Line interface (so it does not disturb the receive pins in High-Z mode)
    result = OtxT1E1XmitPinControl(g_hPhysDevs.hLi[i], OTX_FALSE); DemoCheckRc(g_hPhysDevs.hLi[i], result);

  }

}



/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
void DemoEnablePhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1;

  // Enable Board
  result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  // Enable Quad E1T1 device if board has this device
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_8_PCI:
  case OTX_DEVICE_THOR_8_PCI_PLUS:
  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
  case OTX_DEVICE_GIMLE_16_PCIE:
    for (nQuadE1T1 = 0; nQuadE1T1 < ((g_demoEnv.coLi+3)/4); nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
    break;
  default:
    // Nothing
    break;
  }

  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

}


/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eRxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eTxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eRxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;    // CCS, CAS_BR;
  pT1CfgData->eTxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eRxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eFrameFormat               = g_demoEnv.eT1FrameFormat;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_TRUE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_TRUE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;
  pT1CfgData->bfClearChannels            = 0x000000; // 0x000000 = No channels are marked as clear channels
}




/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
  
}

#define DEMO_EVENT_DATA_BUF_SIZE 10

/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void  *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    DemoLogf("Event signalled\n");

    do {
      OtxEventDataS sqEventData[DEMO_EVENT_DATA_BUF_SIZE];
      OTX_INT32 coEventsFetched;
      OTX_INT32 i;

      result = OtxDrvGetEventDataEx(g_hEventQueue, sqEventData, DEMO_EVENT_DATA_BUF_SIZE, &coEventsFetched); DemoCheckRc(g_hEventQueue, result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event (one at least)
        //DemoLogf("OtxDrvGetEventDataEx() fetched %d events\n", coEventsFetched);

        assert(coEventsFetched <= coEventsFetched);

        for (i = 0; i < coEventsFetched; i++) {
          // Print the event data.
          // In a real application branch based on the data
          DemoPrintEvent(&sqEventData[i]);
        }

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

  }

  return 0;

}



/*******************************************************************************
  @func
    Prints a Driver Event. If HDLC Data, reads the data.
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    
  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (!strncmp(msgStr, "Invalid", strlen("Invalid"))) {
      DemoLogf("E1/T1[%d] Code is %d\n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}

/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->nBoardType        = OTX_DEVICE_THOR_8_PCI_PLUS;
  envP->eClkSource        = OTX_CLOCK_SOURCE_LOCAL_0;
  envP->coLi              = 8;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->eT1E1Mode         = OTX_T1E1_MODE_E1;
  envP->bReceiveEqualizer = OTX_FALSE;
  envP->eT1LineCode       = OTX_T1_LC_B8ZS;
  envP->eT1FrameFormat    = OTX_T1_FF_ESF;
  envP->eE1LineCode       = OTX_E1_LC_HDB3;
  envP->eE1FrameFormat    = OTX_E1_FF_CRC4_MULTIFRAME;
  envP->eLaw              = OTX_DATA_FORMAT_COMPAND8_A_LAW;


  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'H') {   // Longhaul mode
          envP->bReceiveEqualizer = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == '2') {
            envP->nBoardType = OTX_DEVICE_THOR_PCI;
            envP->coLi = 2;
          } else if (*(argv[i] + 2) == '8') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_PCI;
            envP->coLi = 8;
          } else if (*(argv[i] + 2) == 'P') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_PCI_PLUS;
            envP->coLi = 8;
          } else if (*(argv[i] + 2) == 'G') {
            envP->nBoardType  = OTX_DEVICE_GIMLE_16_PCI_PLUS;
            envP->coLi = 16;
          } else if (*(argv[i] + 2) == '9') {
            envP->nBoardType  = OTX_DEVICE_GIMLE_16_PCIE;
            envP->coLi = 16;
          } else if (*(argv[i] + 2) == '4') {
            envP->nBoardType  = OTX_DEVICE_THOR_4_PCI_PLUS;
            envP->coLi = 4;
          } else if (*(argv[i] + 2) == 'T') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_PCI_PLUS;
            envP->coLi = 2;
          } else if (*(argv[i] + 2) == 'E') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_PCI_EXPRESS;
            envP->coLi = 2;
          } else if (*(argv[i] + 2) == 'M') {
            envP->nBoardType  = OTX_DEVICE_THOR_PCMCIA;
            envP->coLi = 2;
          } else if (*(argv[i] + 2) == 'S') {
            envP->nBoardType  = OTX_DEVICE_SLEIPNIR_1_PCI;
            envP->coLi = 1;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'F') {   // T1 Frame format
          if ( *(argv[i]+2) == 'E') {
            envP->eT1FrameFormat = OTX_T1_FF_ESF; // ESF
          } else if ( *(argv[i]+2) == '4') {
            envP->eT1FrameFormat = OTX_T1_FF_F4; 
          } else if ( *(argv[i]+2) == 'S') {
            envP->eT1FrameFormat = OTX_T1_FF_F12; // SF
          } else if ( *(argv[i]+2) == '9') {
            envP->eT1FrameFormat = OTX_T1_FF_F72; // SLC96
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'J') {   // T1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eT1LineCode = OTX_T1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'B') {
            envP->eT1LineCode = OTX_T1_LC_B8ZS; // B8ZS
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'G') {   // E1 Frame format
          if ( *(argv[i]+2) == 'D') {
            envP->eE1FrameFormat = OTX_E1_FF_DOUBLE_FRAME; 
          } else if ( *(argv[i]+2) == 'C') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME; 
          } else if ( *(argv[i]+2) == 'G') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME_G706; 
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'U') {   // E1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eE1LineCode = OTX_E1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'H') {
            envP->eE1LineCode = OTX_E1_LC_HDB3; // HDB3
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'K') {   // Clock Source
          if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '0') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_10;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '1') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_11;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '2') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_12;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '3') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_13;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '4') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_14;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '5') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_15;
          } else if (*(argv[i]+2) == 'I') {   // Internal
            envP->eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
          } else if (*(argv[i]+2) == '0') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_0;
          } else if (*(argv[i]+2) == '1') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_1;
          } else if (*(argv[i]+2) == '2') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_2;
          } else if (*(argv[i]+2) == '3') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_3;
          } else if (*(argv[i]+2) == '4') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_4;
          } else if (*(argv[i]+2) == '5') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_5;
          } else if (*(argv[i]+2) == '6') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_6;
          } else if (*(argv[i]+2) == '7') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_7;
          } else if (*(argv[i]+2) == '8') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_8;
          } else if (*(argv[i]+2) == '9') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_9;
          } else {
            DemoLogf("Error: Invalid Clock Source command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }

        } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {   // E1 
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
            envP->eLaw         = OTX_DATA_FORMAT_COMPAND8_A_LAW;
          } else if (*(argv[i]+2) == 'T') {   // T1
            envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
            envP->eLaw         = OTX_DATA_FORMAT_COMPAND8_U_LAW;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
				  exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
			  exit(1);
      }
    }
  }


}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>    : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[2|T|E|8|P|4|G|9|M|S] : Board Type used for the demo, where\n");
  DemoLogf("                    2  = Use Thor-2-PCI Board\n");
  DemoLogf("                    T  = Use Thor-2-PCI-Plus Board\n");
  DemoLogf("                    E  = Use Thor-2-PCI-Express Board\n");
  DemoLogf("                    8  = Use Thor-8-PCI Board\n");
  DemoLogf("                    P  = Use Thor-8-PCI-Plus Board (default)\n");
  DemoLogf("                    4  = Use Thor-4-PCI-Plus Board\n");
  DemoLogf("                    G  = Use Gimle-16-PCI-Plus Board\n");
  DemoLogf("                    9  = Use Gimle-16-PCIe Board\n");
  DemoLogf("                    M  = Use Thor-2-PCMCIA Board\n");
  DemoLogf("                    S  = Use Sleipnir-1-PCI Board\n");
  DemoLogf(" /L<fileName>   : Log file name\n");
  DemoLogf(" /ME            :   E1 Li mode\n");
  DemoLogf(" /MT            :   T1 Li mode\n");
  DemoLogf(" /H             : Longhaul mode (default is shorthaul mode)\n");
  DemoLogf(" /F[E|S|9]      : T1 Frame Format, where\n");
  DemoLogf("                    E = ESF (default)\n");
  DemoLogf("                    4 = F4\n");
  DemoLogf("                    S = SF [D3/D4] (F12)\n");
  DemoLogf("                    9 = SLC96 (F72)\n");
  DemoLogf(" /J[A|B]        : T1 Line Code, where\n");
  DemoLogf("                    A = AMI\n");
  DemoLogf("                    B = B8ZS (default)\n");
  DemoLogf(" /G[D|C|G]      : E1 Frame Format, where\n");
  DemoLogf("                    D = Doubleframe\n");
  DemoLogf("                    C = CRC4 Multiframe (default)\n");
  DemoLogf("                    G = CRC4 Multiframe (G.706)\n");
  DemoLogf(" /U[A|H]        : E1 Line Code, where\n");
  DemoLogf("                    A = AMI\n");
  DemoLogf("                    H = HDB3 (default)\n");
  DemoLogf(" /K[I|x]        : Clock Source, where\n");
  DemoLogf("                   x = Li number\n");
  DemoLogf(" /Z<str>        : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q             : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(const char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF((char *) formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

}


/*******************************************************************************

  @func

  Displays a message on the screen 

  @comm

*******************************************************************************/
static void DemoPrintf(const char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF((char *) formatStr, argList);

  va_end(argList);

}

static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  static OTX_INT32      nCurrentLi = 0;
  OTX_RESULT result;

  DemoLogf("\nKey %c pressed\n", commandChar);

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" n   Increment current Li (currently Li#%d selected\n", nCurrentLi);
    DemoLogf(" N   Decrement current Li (currently Li#%d selected\n", nCurrentLi);
    DemoLogf(" l   Query Line Status of all Lis\n");
    DemoLogf(" L   Query Extended Line Status of all Lis\n");
    DemoLogf(" F   Query T1/E1 span frequency of the current clock source\n");
    DemoLogf(" z   Set T1/E1 Line Termination for current Li\n");
    DemoLogf(" &   Read receive equalizer status (T1 mode only)\n");
    DemoLogf(" J   Set Receive Input Threshold for current Li\n");
    DemoLogf(" m/M Enable/Disable Monitor mode for current Li (for -20dB attenuation)\n");
    DemoLogf(" g/G Turn front end Amplifier On/Off for current Li (for -30dB attenuation)\n");
    DemoLogf(" f   Force Resynchronization on all Lis\n");
    DemoLogf(" s   Connect HS0 and HS1 to E1/T1 span\n");
    DemoLogf(" c   Set Clock Source\n");
    break;
#if 1 
  case 'x':
    {
      DemoToggleT1E1XmitPinControl();
    }
    break;
#endif
  case 'f':
    {
      DemoLogf("Forcing resynchronization on Li #%d... ", nCurrentLi);
      result = OtxT1E1ForceResynch(g_hPhysDevs.hLi[nCurrentLi]); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      DemoLogf("Done.");
    }
    break;

  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Line status of Li #%d is ", nLi);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[nLi], &eLayer1Status); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }
    }
    break;

  case 'L':
    {
      OtxT1E1Layer1StatusExS sLayer1Status;
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Li #%d line is ", nLi);
        result = OtxT1E1GetLayer1StatusEx(g_hPhysDevs.hLi[nLi], &sLayer1Status); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);

        DemoLogf("Alarms: ");
        if (sLayer1Status.bRedAlarm) {
          DemoLogf("R");
        } else {
          DemoLogf("-");
        }
        if (sLayer1Status.bBlueAlarm) {
          DemoLogf("B");
        } else {
          DemoLogf("-");
        }
        if (sLayer1Status.bYellowAlarm) {
          DemoLogf("Y");
        } else {
          DemoLogf("-");
        }

        if (sLayer1Status.bFrameError) {
          DemoLogf(" FrmErr: YES");
        } else {
          DemoLogf(" FrmErr: NO ");
        }

        if (sLayer1Status.bMultiFrameError) {
          DemoLogf(" MFrmErr: YES");
        } else {
          DemoLogf(" MFrmErr: NO ");
        }
        DemoLogf("\n");
      }
    }
    break;


  case 'm':
    DemoLogf("Enabling Monitor (-20dB) mode for Li #%d \n", nCurrentLi);
    result = OtxT1E1SetMonitorMode(g_hPhysDevs.hLi[nCurrentLi], OTX_TRUE); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
    DemoLogf("Done.\n");
    break;

  case 'M':
    DemoLogf("Disabling Monitor (-20dB) mode for Li #%d \n", nCurrentLi);
    result = OtxT1E1SetMonitorMode(g_hPhysDevs.hLi[nCurrentLi], OTX_FALSE); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
    DemoLogf("Done.\n");
    break;

  case 'z':
    {
      OtxT1E1TermModeE eTermMode;
      int c;

      DemoLogf("Set Line Termination mode:\n");
      DemoLogf("0 = OTX_T1E1_TERM_120_OHM\n");
      DemoLogf("1 = OTX_T1E1_TERM_100_OHM\n");
      DemoLogf("2 = OTX_T1E1_TERM_75_OHM\n");
      DemoLogf("3 = OTX_T1E1_TERM_HIGHZ\n");
      DemoLogf("4 = OTX_T1E1_TERM_300_OHM (for highly attenuated signals)\n");
      c = OsaConsoleGetCh();
      DemoLogf("'%c' pressed\n", c);
      switch (c - '0') {
      case 0:
        eTermMode = OTX_T1E1_TERM_120_OHM;
        break;
      case 1:
        eTermMode = OTX_T1E1_TERM_100_OHM;
        break;
      case 2:
        eTermMode = OTX_T1E1_TERM_75_OHM;
        break;
      case 3:
        eTermMode = OTX_T1E1_TERM_HIGHZ;
        break;
      case 4:
        eTermMode = OTX_T1E1_TERM_300_OHM;
        break;
      default:
        DemoLogf("Invalid entry. Setting Line Termination to OTX_T1E1_TERM_100_OHM\n");
        eTermMode = OTX_T1E1_TERM_100_OHM;
        break;
      }

      result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[nCurrentLi], eTermMode); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Done.\n");
    }
    break;

  case 'g':
    {
      DemoLogf("Turning front end Amplifier ON for Li%d \n", nCurrentLi);
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_RX_AMPLIFIER, OTX_TRUE); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      DemoLogf("Done.\n");
    }
    break;

  case 'G':
    {
      DemoLogf("Turning front end Amplifier OFF for Li%d \n", nCurrentLi);
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_RX_AMPLIFIER, OTX_FALSE); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
      DemoLogf("Done.\n");
    }
    break;


  case 'F':
    {
      OTX_UINT32 nFrequencyCount;

      DemoLogf("Querying frequency of current clock source ...\n");

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_FREQUENCY_COUNTER, &nFrequencyCount); DemoCheckRc(g_hPhysDevs.hThorController0, result);

      if (OTX_SUCCESS(result)) {
        OTX_UINT32 nFrequency;

        if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
          // T1 mode
          nFrequency = (nFrequencyCount+1)*193/2048;
        } else {
          // E1 mode
          nFrequency = (nFrequencyCount+1)/8;
        }
        DemoLogf("Line frequency is %d Hz (nFrequencyCount=%d)\n", nFrequency, nFrequencyCount);
      }

      DemoLogf("Done.\n");
    }
    break;

  case 'r':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, &readVal); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Li%d Address 0x%X = %.2X\n", nCurrentLi, nAddress, (OTX_UINT8)readVal);
    }
    break;

  case 'w':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, nValue); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Wrote %.2X to Li%d Address 0x%X\n", (OTX_UINT8)nValue, nCurrentLi, nAddress);
    }
    break;

  case 'c':
    {
      OtxBrdClockSourceE  eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
      OTX_INT32 c;

      DemoLogf("Set Clock Source:\n");
      if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI || g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS) {
        printf("0 = OTX_CLOCK_SOURCE_LOCAL_0  - LIs synchronized to Li0 receive span (QuadE1T1#0)\n");
        printf("1 = OTX_CLOCK_SOURCE_LOCAL_1  - LIs synchronized to Li1 receive span (QuadE1T1#0)\n");
        printf("2 = OTX_CLOCK_SOURCE_LOCAL_2  - LIs synchronized to Li2 receive span (QuadE1T1#0)\n");
        printf("3 = OTX_CLOCK_SOURCE_LOCAL_3  - LIs synchronized to Li3 receive span (QuadE1T1#0)\n");
        printf("4 = OTX_CLOCK_SOURCE_LOCAL_4  - LIs synchronized to Li4 receive span (QuadE1T1#1)\n");
        printf("5 = OTX_CLOCK_SOURCE_LOCAL_5  - LIs synchronized to Li5 receive span (QuadE1T1#1)\n");
        printf("6 = OTX_CLOCK_SOURCE_LOCAL_6  - LIs synchronized to Li6 receive span (QuadE1T1#1)\n");
        printf("7 = OTX_CLOCK_SOURCE_LOCAL_7  - LIs synchronized to Li7 receive span (QuadE1T1#1)\n");
        printf("i = OTX_CLOCK_SOURCE_INTERNAL - LIs synchronized to on-board oscillator\n");

        c = OsaConsoleGetCh();
        DemoLogf("'%c' pressed\n", c);
        if (c == 'i') {
          eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
        } else {
          switch (c - '0') {
          case 0:
          case 1:
          case 2:
          case 3:
          case 4:
          case 5:
          case 6:
          case 7:
            eClkSource = (OtxBrdClockSourceE) ( OTX_CLOCK_SOURCE_LOCAL_0 + (c - '0') ); 
            break;
          default:
            printf("Invalid entry. Setting Clock Source to OTX_CLOCK_SOURCE_INTERNAL\n");
            eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
            break;
          }
        }
      } else if (g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_PLUS || g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCIE) {
        printf("0 = OTX_CLOCK_SOURCE_LOCAL_0  - LIs synchronized to Li0 receive span  (QuadE1T1#0)\n");
        printf("1 = OTX_CLOCK_SOURCE_LOCAL_1  - LIs synchronized to Li1 receive span  (QuadE1T1#0)\n");
        printf("2 = OTX_CLOCK_SOURCE_LOCAL_2  - LIs synchronized to Li2 receive span  (QuadE1T1#0)\n");
        printf("3 = OTX_CLOCK_SOURCE_LOCAL_3  - LIs synchronized to Li3 receive span  (QuadE1T1#0)\n");
        printf("4 = OTX_CLOCK_SOURCE_LOCAL_4  - LIs synchronized to Li4 receive span  (QuadE1T1#1)\n");
        printf("5 = OTX_CLOCK_SOURCE_LOCAL_5  - LIs synchronized to Li5 receive span  (QuadE1T1#1)\n");
        printf("6 = OTX_CLOCK_SOURCE_LOCAL_6  - LIs synchronized to Li6 receive span  (QuadE1T1#1)\n");
        printf("7 = OTX_CLOCK_SOURCE_LOCAL_7  - LIs synchronized to Li7 receive span  (QuadE1T1#1)\n");
        printf("8 = OTX_CLOCK_SOURCE_LOCAL_8  - LIs synchronized to Li8 receive span  (QuadE1T1#2)\n");
        printf("9 = OTX_CLOCK_SOURCE_LOCAL_9  - LIs synchronized to Li9 receive span  (QuadE1T1#2)\n");
        printf("a = OTX_CLOCK_SOURCE_LOCAL_10 - LIs synchronized to Li10 receive span (QuadE1T1#2)\n");
        printf("b = OTX_CLOCK_SOURCE_LOCAL_11 - LIs synchronized to Li11 receive span (QuadE1T1#2)\n");
        printf("c = OTX_CLOCK_SOURCE_LOCAL_12 - LIs synchronized to Li12 receive span (QuadE1T1#3)\n");
        printf("d = OTX_CLOCK_SOURCE_LOCAL_13 - LIs synchronized to Li13 receive span (QuadE1T1#3)\n");
        printf("e = OTX_CLOCK_SOURCE_LOCAL_14 - LIs synchronized to Li14 receive span (QuadE1T1#3)\n");
        printf("f = OTX_CLOCK_SOURCE_LOCAL_15 - LIs synchronized to Li15 receive span (QuadE1T1#3)\n");
        printf("i = OTX_CLOCK_SOURCE_INTERNAL - LIs synchronized to on-board oscillator\n");

        c = OsaConsoleGetCh();
        DemoLogf("'%c' pressed\n", c);
        if (c == 'i') {
          eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
        } else if (c >= '0' && c <= '9') {
          eClkSource = (OtxBrdClockSourceE) ( OTX_CLOCK_SOURCE_LOCAL_0 + (c - '0') ); 
        } else if (c >= 'a' && c <= 'f') {
          eClkSource = (OtxBrdClockSourceE) ( OTX_CLOCK_SOURCE_LOCAL_10 + (c - 'a') ); 
        }
      } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_PCI_PLUS) {
        printf("0 = OTX_CLOCK_SOURCE_LOCAL_0  - LIs synchronized to Li0 receive span (QuadE1T1#0)\n");
        printf("1 = OTX_CLOCK_SOURCE_LOCAL_1  - LIs synchronized to Li1 receive span (QuadE1T1#0)\n");
        printf("2 = OTX_CLOCK_SOURCE_LOCAL_2  - LIs synchronized to Li2 receive span (QuadE1T1#0)\n");
        printf("3 = OTX_CLOCK_SOURCE_LOCAL_3  - LIs synchronized to Li3 receive span (QuadE1T1#0)\n");
        printf("i = OTX_CLOCK_SOURCE_INTERNAL - LIs synchronized to on-board oscillator\n");

        c = OsaConsoleGetCh();
        DemoLogf("'%c' pressed\n", c);
        if (c == 'i') {
          eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
        } else {
          switch (c - '0') {
          case 0:
          case 1:
          case 2:
          case 3:
            eClkSource = (OtxBrdClockSourceE) ( OTX_CLOCK_SOURCE_LOCAL_0 + (c - '0') ); 
            break;
          default:
            printf("Invalid entry. Setting Clock Source to OTX_CLOCK_SOURCE_INTERNAL\n");
            eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
            break;
          }
        }
      } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA) {
        DemoLogf("0 = LOCAL_0  - LIs synchronized to Li0 receive span\n");
        DemoLogf("1 = LOCAL_1  - LIs synchronized to Li1 receive span\n");
        DemoLogf("2 = LOCAL_2  - External clock source\n");
        DemoLogf("3 = LOCAL_3  - Split Clock, both Line Interface are synched to its T1E1 span\n");
        DemoLogf("4 = LOCAL_4  - Split Clock, Li0 synchronized to its T1E1 span. Li1 internal\n");
        DemoLogf("5 = LOCAL_5  - Split Clock, Li1 synchronized to its T1E1 span. Li0 internal\n");
        DemoLogf("i = INTERNAL - LIs synchronized to on-board oscillator\n");

        c = OsaConsoleGetCh();
        DemoLogf("'%c' pressed\n", c);
        if (c == 'i') {
          eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
        } else {
          switch (c - '0') {
          case 0:
          case 1:
          case 2:
          case 3:
          case 4:
          case 5:
            eClkSource = (OtxBrdClockSourceE) (OTX_CLOCK_SOURCE_LOCAL_0 + (c - '0') );
            break;
          default:
            DemoLogf("Invalid entry. Setting Clock Source to OTX_CLOCK_SOURCE_INTERNAL\n");
            eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
            break;
          }
        }
      } else {
        DemoLogf("0 = OTX_CLOCK_SOURCE_LOCAL_0  - LIs synchronized to Li0 receive span\n");
        DemoLogf("1 = OTX_CLOCK_SOURCE_LOCAL_1  - LIs synchronized to Li1 receive span\n");
        DemoLogf("i = OTX_CLOCK_SOURCE_INTERNAL - LIs synchronized to on-board oscillator\n");

        c = OsaConsoleGetCh();
        DemoLogf("'%c' pressed\n", c);
        if (c == 'i') {
          eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
        } else {
          switch (c - '0') {
          case 0:
          case 1:
            eClkSource = (OtxBrdClockSourceE) (OTX_CLOCK_SOURCE_LOCAL_0 + (c - '0') );
            break;
          default:
            DemoLogf("Invalid entry. Setting Clock Source to OTX_CLOCK_SOURCE_INTERNAL\n");
            eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
            break;
          }
        }

      }

      DemoLogf("Calling OtxBrdSetClocks(%d)\n", (int)eClkSource);
      result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, eClkSource); DemoCheckRc(g_hPhysDevs.hThorController0, result);

      DemoLogf("Done.\n");
    }
    break;
  

  case '&':
    {
      OTX_UINT32 readVal;
      OTX_INT32 nLi;
      const char *pszEqStatus;
      
      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        result = OtxDrvReadReg8(g_hPhysDevs.hLi[nLi], 115, &readVal); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
        switch ((readVal & 0xC0) >> 6) {
        case 0x00:
          pszEqStatus = "not valid, still adapting";
          break;
        case 0x01:
          pszEqStatus = "valid";
          break;
        case 0x02:
          pszEqStatus = "not valid";
          break;
        case 0x03:
          pszEqStatus = "valid but high noise floor";
          break;
        default:
          pszEqStatus = "unknown";
          break;
        }
        DemoLogf("Receive Equalizer of Li#%d: status=%s attenuation=%2.1fdB (0x%.2X)\n", nLi, pszEqStatus, (readVal & 0x1F)*1.4, readVal );
      }
      
    }
    break; 
    
  case 'J':
    {
      char strBuf[12];
      OTX_UINT32 nThresholdInMilliVolts;

      DemoLogf("Set Receive Input Threshold for Li%d \n", nCurrentLi);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, &nThresholdInMilliVolts); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Threshold (%d mV): ", nThresholdInMilliVolts);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nThresholdInMilliVolts = strtoul(strBuf, NULL, 0);

        result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, nThresholdInMilliVolts); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

        result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, &nThresholdInMilliVolts); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
        DemoLogf("Receive Input Threshold is changed to %d mV.\n", nThresholdInMilliVolts);
      } else {
        DemoLogf("The Receive Input Threshold remains unchanged.\n");
      }

    }
    break;

  case 's':
    {
      OTX_INT32 i;
      OTX_INT32 nTs = 1;
      char strBuf[12];
      OTX_UINT8 sqTimeSlotMask[32];
      static OTX_HANDLE hLiCodecPipe[2] = {OTX_INVALID_HANDLE_VALUE, OTX_INVALID_HANDLE_VALUE}; 
  
      for (i = 0; i < 2; i++) {
        if (hLiCodecPipe[i] == OTX_INVALID_HANDLE_VALUE) {
          result = OtxDrvCreatePipe(64, &hLiCodecPipe[i]); DemoCheckRc(hLiCodecPipe[i], result);

          result = OtxDrvConnectPipe(hLiCodecPipe[i], OTX_PIPE_OUTPUT, g_hPhysDevs.hCodec[i], -1, NULL, 0); DemoCheckRc(hLiCodecPipe[i], result);
        }
      }
        
      DemoLogf("Connect HS0 to Li0 and Hs1 to Li1. Specify TS: ");

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nTs = strtol(strBuf, NULL, 0);
      }

      if (nTs < 1 || nTs > 31) {
        DemoLogf("TS must be in the range of 1-31. Operation aborted.\n");
        break;
      }

      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[nTs] = 0xFF;
        
      for (i = 0; i < 2; i++) {
        // In case the input was connected earlier, disconnect it; 
        OtxDrvDisconnectPipe(hLiCodecPipe[i], OTX_PIPE_INPUT);

        result = OtxDrvConnectPipe(hLiCodecPipe[i], OTX_PIPE_INPUT, g_hPhysDevs.hLi[i], -1, sqTimeSlotMask, 32); DemoCheckRc(hLiCodecPipe[i], result);
      }

      DemoLogf("HS0 and HS1 are now connected to TS%d of Li0 and Li1\n", nTs);
        
    }

    break;



  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoPrintf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoPrintf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  case 'n':
    {
      // Select current nCurrentLi
      nCurrentLi++;

      if (nCurrentLi >= g_demoEnv.coLi) {
        // Wrap 
        nCurrentLi = 0;
      }
      
      DemoLogf("Li#%d selected\n", nCurrentLi);
      
    }
    break;
    
    
  case 'N':
    {
      // Select current nCurrentLi
      nCurrentLi--;

      if (nCurrentLi < 0) {
        // Wrap
        nCurrentLi = g_demoEnv.coLi - 1;
      }
      
      DemoLogf("Li#%d selected \n", nCurrentLi);
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}


/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  fclose(g_pDemoLogFile);
}


/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  DemoLogf("LineMon, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("E1T1 Monitoring Example program.\n");
  DemoLogf("For OTX Thor, Gimle and Sleipnir boards\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }
  atexit(DemoAtExit);

  DemoLogf("Log file is  %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }

  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  // Query driver revision
  {
    char szVersionInfo[255];
    result = OtxDrvGetAttributeValueSz(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Driver version %s\n", szVersionInfo);
  }

  // Create Event Queues for Driver --> Application Communications
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the board
  DemoOpenPhysicalDevices();

  // Configure E1/T1 devices
  DemoConfigureE1T1Devices();

  // Enable all physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoEnablePhysicalDevices();

  {
    char szClockSource[64];

    switch (g_demoEnv.eClkSource) {
    case OTX_CLOCK_SOURCE_INTERNAL:
      strcpy(szClockSource, "Internal");
      break;
    case OTX_CLOCK_SOURCE_LOCAL_0:
    case OTX_CLOCK_SOURCE_LOCAL_1:
    case OTX_CLOCK_SOURCE_LOCAL_2:
    case OTX_CLOCK_SOURCE_LOCAL_3:
    case OTX_CLOCK_SOURCE_LOCAL_4:
    case OTX_CLOCK_SOURCE_LOCAL_5:
    case OTX_CLOCK_SOURCE_LOCAL_6:
    case OTX_CLOCK_SOURCE_LOCAL_7:
    case OTX_CLOCK_SOURCE_LOCAL_8:
    case OTX_CLOCK_SOURCE_LOCAL_9:
    case OTX_CLOCK_SOURCE_LOCAL_10:
    case OTX_CLOCK_SOURCE_LOCAL_11:
    case OTX_CLOCK_SOURCE_LOCAL_12:
    case OTX_CLOCK_SOURCE_LOCAL_13:
    case OTX_CLOCK_SOURCE_LOCAL_14:
    case OTX_CLOCK_SOURCE_LOCAL_15:
      sprintf(szClockSource, "Line Interface %d", (int)g_demoEnv.eClkSource - (int)OTX_CLOCK_SOURCE_LOCAL_0);
      break;
    default:
      strcpy(szClockSource, "Unknown");
      break;
    }

    DemoLogf("Clock Source: %s\n", szClockSource);
  }

  result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, g_demoEnv.eClkSource); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {
    // Read CPLD Build Numbers from the board
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszCpldBuildNumbers;
    
    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);
    free(pszCpldBuildNumbers);
  }

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....

  OsaThreadStart(DemoIdleFunction);

  OsaThreadSleep(300);

  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OtxDrvReleaseLib();
  OsaConsoleExit();

  return 0;
}

