/* $Workfile: SingleTone.c $       $Revision: 2 $      $Modtime: 10/24/12 12:55p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    SingleTone.c

    A simple example console application demonstrating the use the following
    devices:
    
      OTX_LDEVICE_TONE_SINGLE_DETECTOR, 
      OTX_LDEVICE_TONE_SINEWAVE_GENERATOR,
    
    The OTX_LDEVICE_TONE_SINEWAVE_GENERATOR device generates the following
    telephony tones (U.S. standard)

        Ring Tone
        Busy Tone
        Dial Tone

    The OTX_LDEVICE_TONE_SINGLE_DETECTOR device is configured to detect these
    tones.

    The applications works with the following board combinations:

    OTX PCI Plus series boards,

      Thor-2[4,8]-PCI-Plus + Vidar-5x4-ASM[-CST,-PRO,-EX]
      Thor-2[4,8]-PCI-Plus + Vidar-55x4-ASM,
      Thor-2[4,8]-PCI-Plus + Alvis-ASM,

    or OTX PCI Express series boards,

      Thor-2-PCI-Express + Vidar-5x4-ASM[-CST,-PRO,-EX]
      Thor-2-PCI-Express + Vidar-55x4-ASM,
      Thor-8-PCIe + Vidar-55x4-ASM,

    or Thor-2-PCMCIA series boards,

      Thor-2-PCMCIA[-Plus,-EX,-PRO,-CST] 

    or Sleipnir-1-PCI-Plus

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures 2 E1 Lines and expects them to be looped with an external
    loop cable. 

    The tone devices are configured in two DSPs (one for each LI) using the
    OTXSPM2 DSP module.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <assert.h>
#include <time.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxTone.h"
#include "OtxDsp.h"
#include "OtxHdlc.h"
#include "OtxT1E1.h"
#include "OtxData.h"

#define DEMO_TASK_REF 1L               // Task Reference for a non-blocking call

#define DEMO_DEFAULT_LOG_FILE "SingleTone.log"
#define DEMO_DEFAULT_BOARD_NO 0

typedef struct DemoEnvT {
  char            *pszInputStr;
  OTX_BOOL        bExitOnError;
  OTX_INT32       nBoardType;
  char            cwd[OTX_MAX_PATH];
  char            logFileName[OTX_MAX_PATH];
  OtxDataFormatE  law;
  OtxT1E1LiModeE  eT1E1Mode;    // T1 or E1 mode
  int             boardNo;
  int             coLi;
  int             coDsp;
  OtxDspSpmE      eProgramModule;      // Standard Program Module to be loaded in the DSP (from the Lib)
  int             nLinkTimeout;
  char           *pszDspOutFile;   // Optionally load the DSP program direct from an OUT file (this file name)
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoConfigureE1T1Devices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoOpenPhysicalDevices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);

static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(const char *formatStr, ...);
static void DemoCheckRc(IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
static void DemoPrintTimeStamp(OTX_DATETIME nTimeStamp);

static OTX_INT32 DemoProcessUserCommand(char commandChar);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_UINT32  g_nDspHighwaySpeed;   // DSP Highway speed in kpbs
OTX_UINT32  g_eDevType = OTX_DEVICE_ALVIS_ASM;

struct PhysDevs {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hQuadE1T1[2]; // Only used for Thor-8-PCI; not for Thor-2-PCI
  OTX_HANDLE hVidarAsmController0;
  OTX_HANDLE hLi[8];
  OTX_HANDLE hDsp[4];
  OTX_HANDLE hEthSw;
} g_hPhysDevs;


struct Pipes {
  OTX_HANDLE hLiDspSingleToneDetector[4];
  OTX_HANDLE hDspLiSineWaveSender[4];
  OTX_HANDLE hLiIn[2];
  OTX_HANDLE hLiOut[2];
} g_hPipes;


struct LogicDevs {
  OTX_HANDLE hSingleToneDetector[4];
  OTX_HANDLE hSineWaveSender[4];
} g_hLogicDevs;


struct T1E1CfgData {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;


/*******************************************************************************
*******************************************************************************/


/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DTMF digits.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;
    
  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = (OTX_EVENT) NULL;    // D16V86 apps cannot use event objects
    result = OTX_S_OK;
  }
    
  DemoCheckRc(result);						
    
  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(result);
    
  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(result);
}


/*******************************************************************************

  @func
    Open the Physical Devices on the board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
static void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nDsp;
  OTX_INT32 nQuadE1T1; // Only used for Thor-4-PCI-Plus, Thor-8-PCI, Thor-8-PCI-Plus
    
  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_2_PCI_PLUS:
    DemoLogf("Opening Thor-2-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    DemoLogf("Opening Thor-2-PCI-Express board\n");
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    DemoLogf("Opening Thor-4-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_PCMCIA:
    DemoLogf("Opening Thor-2-PCMCIA variant board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    DemoLogf("Opening Thor-8-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_8_PCIE:
    DemoLogf("Opening Thor-8-PCIe board\n");
    break;
  case OTX_DEVICE_SLEIPNIR_1_PCI:
    DemoLogf("Opening Sleipnir-1-PCI board\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
    
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE)g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController0); DemoCheckRc(result);
    
  // Get handles to T1/E1 line interface devices #0 and #1
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCMCIA:
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
  case OTX_DEVICE_SLEIPNIR_1_PCI:
    // On these boards the E1T1 devices are hosted by the board itself. Get handles to T1/E1 line interface devices #0 and #1
    for (i = 0; i < g_demoEnv.coLi; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
    }
    break;
    
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    {
      // On the Thor-4-PCI-Plus board the E1T1 devices are hosted by a quad E1T1 device. 
      for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < g_demoEnv.coLi; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
      }
    }
    break;

  case OTX_DEVICE_THOR_8_PCI_PLUS:
  case OTX_DEVICE_THOR_8_PCIE:
    {        
      // On the Thor-8-PCI board the E1T1 devices are hosted by a quad E1T1 device. Open those two devices.
      for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(result);
      }
        
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < g_demoEnv.coLi; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
      }
    }
    break;
    
  default:
      DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
      exit(1);
      break;
  }
    
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_SLEIPNIR_1_PCI:
  case OTX_DEVICE_THOR_PCMCIA:    
    // Get handles to the physcial DSP devices #0 and/or #1
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(result);
    }
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
  case OTX_DEVICE_THOR_8_PCIE:
  case OTX_DEVICE_THOR_4_PCI_PLUS:
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    {    
      // Get a handle to to the ASM Daughter Board
      OTX_UINT32   eDevType;
      OTX_ITERATOR itrChildTypes;
      
      OtxDrvIteratorBegin(&itrChildTypes);
    
      while (OTX_SUCCESS(OtxDrvEnumDeviceType(g_hPhysDevs.hThorController0, &itrChildTypes, (OTX_INT32 *)&eDevType))) {  
        OTX_BOOL bOpenAsm = OTX_FALSE;
        switch (eDevType) {
        case OTX_DEVICE_VIDAR_5x_ASM:
          DemoLogf("Opening Vidar-5x4-ASM board\n");
          bOpenAsm = OTX_TRUE;
          break;
        case OTX_DEVICE_VIDAR_5X4_ASM_CST:
          DemoLogf("Opening Vidar-5x4-ASM-CST board\n");
          bOpenAsm = OTX_TRUE;
          break;
        case OTX_DEVICE_VIDAR_5X4_ASM_V90:
          DemoLogf("Opening Vidar-5x4-ASM-V90 board\n");
          bOpenAsm = OTX_TRUE;
          break;
        case OTX_DEVICE_VIDAR_5X4_ASM_EX:
          DemoLogf("Opening Vidar-5x4-ASM-EX board\n");
          bOpenAsm = OTX_TRUE;
          break;
        case OTX_DEVICE_VIDAR_5X4_ASM_PRO:
          DemoLogf("Opening Vidar-5x4-ASM-PRO board\n");
          bOpenAsm = OTX_TRUE;
          break;
        case OTX_DEVICE_VIDAR_55x4_ASM:
          DemoLogf("Opening Vidar-55x4-ASM board\n");
          bOpenAsm = OTX_TRUE;
          break;

        case OTX_DEVICE_ALVIS_ASM:
          DemoLogf("Alvis-ASM board is not supported by this demo\n");
          g_eDevType = OTX_DEVICE_ALVIS_ASM;
          return;

        }
      
        if (bOpenAsm) {
          // Get a handle to to the ASM Daughter Board
          result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, (OtxPhysicalDeviceTypeE) eDevType, 0, 0, g_hEventQueue, &g_hPhysDevs.hVidarAsmController0); DemoCheckRc(result);
        }      
      } //while
      
      for (i = 0; i < g_demoEnv.coDsp; i++) {
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hVidarAsmController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(result);
      } 
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }

  
  // Load DSP Program
  for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
    
    switch (g_demoEnv.eProgramModule) {
    case OTX_DSP_SPM_OTXSPM2:
      DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXSPM2\n", nDsp);
      break;
    case OTX_DSP_SPM_OTXCST3:
      DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXCST3\n", nDsp);
      break;
    default:
      DemoLogf("Unsupport DSP program module (%d) for this demo program. Exiting.\n", g_demoEnv.eProgramModule);
      OsaConsoleExit();
      OtxDrvReleaseLib();
      exit(1);
    }
    
    result = OtxDspRunSpm(g_hPhysDevs.hDsp[nDsp], g_demoEnv.eProgramModule); DemoCheckRc(result);
    
  }
    
}


/*******************************************************************************

  @func
    Configure E1/T1 Devices

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
    
  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));
    
  // Configure Transceivers
  for (i = 0; i < g_demoEnv.coLi; i++) {
      
    // Set termination for the T1/E1 Front End
    result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(result);
      
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(result);
    } else {
      // E1 mode
      result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(result);
    }
    
    result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(result);
  }
}

  
/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eRxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = OTX_FALSE;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;
  pT1CfgData->eTxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eRxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eFrameFormat               = OTX_T1_FF_ESF;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = OTX_TRUE;
  pT1CfgData->bfClearChannels            = 0x000000; // 0x000000 = No channels are marked as clear channels
}


/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
void DemoEnablePhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
    
  result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  if (g_hPhysDevs.hVidarAsmController0) {
    result = OtxDrvEnable(g_hPhysDevs.hVidarAsmController0, OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }
    
  // If board type is Thor-8-PCI then enable Quad E1T1 device
  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS || g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCIE) {
    OTX_INT32 nQuadE1T1; // Only used for Thor-8-PCI
      
    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    }
  }
       
  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hLi[i],            OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }
    
  for (i = 0; i < g_demoEnv.coDsp; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i],            OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }
}


/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

*******************************************************************************/
void DemoCreateLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;
    
  for (i = 0; i < g_demoEnv.coDsp; i++) {

    // Create Tone Detector Logical Devices
    switch (g_demoEnv.eProgramModule) {
    case OTX_DSP_SPM_OTXSPM2:
      result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[i], OTX_LDEVICE_TONE_SINGLE_DETECTOR, g_hEventQueue, i, &g_hLogicDevs.hSingleToneDetector[i]); DemoCheckRc(result);
      break;
    case OTX_DSP_SPM_OTXCST3:
      result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[i], OTX_LDEVICE_CST_CPT_DETECTOR, g_hEventQueue, i, &g_hLogicDevs.hSingleToneDetector[i]); DemoCheckRc(result);
      break;
    default:
      DemoLogf("Unsupport DSP program module (%d) for this demo program. Exiting.\n", g_demoEnv.eProgramModule);
      OsaConsoleExit();
      OtxDrvReleaseLib();
      exit(1);
    }


      
    // Create Sinewave Sender Logical Devices
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_TONE_SINEWAVE_GENERATOR, g_hEventQueue, i, &g_hLogicDevs.hSineWaveSender[i]); DemoCheckRc(result);      
  }    
}


/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(
  void
)
{
  OTX_INT32 nLi;
  OTX_INT32 nDsp;
  OTX_INT32 i;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_UINT8 sqDspTimeSlotMask[128];
  OTX_INT32 ts;
  OTX_UINT32 coTsPerDspHwy = 0; // used if g_eDevType == OTX_DEVICE_ALVIS_ASM
  OTX_UINT32 nDspHighwaySpeed;
  OTX_INT32 nDspTs;
    
  // Create pipes
  for (i = 0; i < g_demoEnv.coDsp; i++) {
    result = OtxDrvCreatePipe(64, &g_hPipes.hLiDspSingleToneDetector[i]); DemoCheckRc(result);
    result = OtxDrvCreatePipe(64, &g_hPipes.hDspLiSineWaveSender[i]); DemoCheckRc(result);
  }                  
        
  // Connect pipes for the SineWave Sender and Single Tone Detector
  for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
    // Create one pipes per Li. One pipe to each DSP
    // Li0 Ts5 <=> DSP0
    // Li1 Ts5 <=> DSP1
      
    nDsp = nLi;
    if (nDsp >= g_demoEnv.coDsp) break;
      
    // Connect Pipes
    ts = 5;
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[ts] = 0xFF;
      
    if (g_eDevType == OTX_DEVICE_ALVIS_ASM) {
      
      // Get DSP Highway Speed
      OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_HIGHWAY_SPEED, &nDspHighwaySpeed); 
      DemoLogf("DSP#%d highway speed is %d kbps\n", nDsp, nDspHighwaySpeed);
      coTsPerDspHwy = nDspHighwaySpeed / 64;

      DemoLogf("DSP[%d] physical handle = %X \n", nDsp, g_hPhysDevs.hDsp[nDsp]);
      nDspTs = ts;
    
      memset(sqDspTimeSlotMask, 0, coTsPerDspHwy * sizeof(OTX_UINT8));
      sqDspTimeSlotMask[nDspTs] = 0xFF;
      
      DemoLogf("Connecting Dual Tone device[%d]: nLi=%d, Li-TS =%.2d, nDsp=%d, Dsp-TS \n", nLi, nLi, ts, nDsp, nDspTs);
    }    

    result = OtxDrvConnectPipe(g_hPipes.hLiDspSingleToneDetector[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32); DemoCheckRc(result);
      
    if ( (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA) || (g_demoEnv.nBoardType == OTX_DEVICE_SLEIPNIR_1_PCI) ) {
      result = OtxDrvConnectPipe(g_hPipes.hLiDspSingleToneDetector[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp], -1, sqTimeSlotMask, 32); DemoCheckRc(result);
    } else if (g_eDevType == OTX_DEVICE_ALVIS_ASM) {
      result = OtxDrvConnectPipe(g_hPipes.hLiDspSingleToneDetector[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp], -1, sqDspTimeSlotMask, coTsPerDspHwy); DemoCheckRc(result);
    } else {
      result = OtxDrvConnectPipe(g_hPipes.hLiDspSingleToneDetector[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp], -1, NULL, 0); DemoCheckRc(result);
    }
      
      
    result = OtxDrvConnectPipe(g_hPipes.hDspLiSineWaveSender[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32); DemoCheckRc(result);
      
    if ( (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA) || (g_demoEnv.nBoardType == OTX_DEVICE_SLEIPNIR_1_PCI) ) {
      result = OtxDrvConnectPipe(g_hPipes.hDspLiSineWaveSender[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp], -1, sqTimeSlotMask, 32); DemoCheckRc(result);
    } else if (g_eDevType == OTX_DEVICE_ALVIS_ASM) {
      result = OtxDrvConnectPipe(g_hPipes.hDspLiSineWaveSender[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp], -1, sqDspTimeSlotMask, coTsPerDspHwy); DemoCheckRc(result);
    } else {
      result = OtxDrvConnectPipe(g_hPipes.hDspLiSineWaveSender[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp], -1, NULL, 0); DemoCheckRc(result);
    }
  }
    
  for (i = 0; i < g_demoEnv.coDsp; i++) {
      
    // CONNECT LOGICAL DEVICES WITHIN THE DSP
    // --------------------------------------
      
    // Create the following Connections in receive direction:
    //  Li  ------> SingleToneDetector
    result = OtxDrvConnectLogicalDevice(g_hPipes.hLiDspSingleToneDetector[i], -1, g_hLogicDevs.hSingleToneDetector[i], -1); DemoCheckRc(result);
      
    // Make the following connections in transmit direction:
    //  SineWaveGenerator   -----> Li
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hSineWaveSender[i], -1, g_hPipes.hDspLiSineWaveSender[i], -1); DemoCheckRc(result);
  }
}


/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.

*******************************************************************************/
void DemoInitializeLogicalDevices(
  void
)
{
  
    switch (g_demoEnv.eProgramModule) {
    case OTX_DSP_SPM_OTXSPM2:
      {
        OTX_RESULT result;
        OTX_INT32  i;
        
        {
          OTX_UINT32 nThreshold;
          i = 0;
          
          result = OtxDrvGetAttributeValueUint32(g_hLogicDevs.hSingleToneDetector[i], OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_TONE_ENERGY, &nThreshold); DemoCheckRc(result);
          DemoLogf("Single Tone Detector[%d]: OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_TONE_ENERGY  = %d\n", i, nThreshold);
          
          result = OtxDrvGetAttributeValueUint32(g_hLogicDevs.hSingleToneDetector[i], OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_VARIANCE_ENERGY, &nThreshold); DemoCheckRc(result);
          DemoLogf("Single Tone Detector[%d]: OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_VARIANCE_ENERGY  = %d\n", i, nThreshold);
          
          result = OtxDrvGetAttributeValueUint32(g_hLogicDevs.hSingleToneDetector[i], OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_PAUSE_ENERGY, &nThreshold); DemoCheckRc(result);
          DemoLogf("Single Tone Detector[%d]: OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_PAUSE_ENERGY = %d\n", i, nThreshold);
          
          // Modifing default thresholds
          nThreshold = 1000;
          DemoLogf("Single Tone Detector[%d]: Setting OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_TONE_ENERGY  = %d\n", i, nThreshold);
          result = OtxDrvSetAttributeValueUint32(g_hLogicDevs.hSingleToneDetector[i], OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_TONE_ENERGY, nThreshold); DemoCheckRc(result);
          
          nThreshold = 200;
          DemoLogf("Single Tone Detector[%d]: Setting OTX_ATTR_SINGLE_VARIANCE_DETECTOR_THRESHOLD_TONE_ENERGY  = %d\n", i, nThreshold);
          result = OtxDrvSetAttributeValueUint32(g_hLogicDevs.hSingleToneDetector[i], OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_VARIANCE_ENERGY, nThreshold); DemoCheckRc(result);
          
          nThreshold = 200;
          DemoLogf("Single Tone Detector[%d]: Setting OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_PAUSE_ENERGY = %d\n", i, nThreshold);
          result = OtxDrvSetAttributeValueUint32(g_hLogicDevs.hSingleToneDetector[i], OTX_ATTR_SINGLE_TONE_DETECTOR_THRESHOLD_PAUSE_ENERGY, nThreshold); DemoCheckRc(result);
          
        }
        
        for (i = 0; i < g_demoEnv.coDsp; i++) {
          
          OTX_UINT32    sqFreq[2];       // An array of up to 6 frequencies to be detected.
          OTX_INT32     coFreq = 2;       // Number of elements in <p sqFreq>
          
#if 0 // Include this code to get a more status messages from the Single Tone Detector
          result = OtxDrvSetAttributeValueUint32(g_hLogicDevs.hSingleToneDetector[i], OTX_ATTR_SINGLE_TONE_DETECTOR_VERBOSE_STATUS, 5); DemoCheckRc(result);
#endif
          
          sqFreq[0] = 440;  // Freq A in Hz
          sqFreq[1] = 620;  // Freq B in Hz       
          
          
          // Initialize the Generic Single Tone Detector. Use a couple of DTMF tone frequencies as an example
          result =  OtxToneSingleDetectorInit(
            g_hLogicDevs.hSingleToneDetector[i],           // Handle to the Logical device
            g_demoEnv.law,    // Data Format of the Monitored Data Stream
            sqFreq,           // An array of up to 6 frequencies to be detected.
            coFreq,           // Number of elements in <p sqFreq>
            OTX_TRUE,         // If OTX_TRUE, generate events and store the received digits
            OTX_TRUE,         // If OTX_TRUE the detector generates an event when a tone ends
            250                // Minimum duration (in milliseconds) required to validate a dual tone
            );
          DemoCheckRc(result);
          
        }

      }
      break;

    default:
      DemoLogf("Unsupport DSP program module (%d) for this demo program. Exiting.\n", g_demoEnv.eProgramModule);
      OsaConsoleExit();
      OtxDrvReleaseLib();
      exit(1);
    }


    
}


/*******************************************************************************

  @func
    Enable Configured Logical Devices

*******************************************************************************/
void DemoEnableLogicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
    
  for (i = 0; i < g_demoEnv.coDsp; i++) {
    result = OtxDrvEnable(g_hLogicDevs.hSingleToneDetector[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  }    
}


/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*********************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;


  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

  return 0;


}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
    
}


/*******************************************************************************
  @func
    Prints a Driver Event. If HDLC Data, reads the data.
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  //OTX_DATETIME nTimeStamp;
  char msgStr[256];

  //DemoLogf("%.8X%.8X ", pEventData->m_nTimeStamp.sUint32.nHigh, pEventData->m_nTimeStamp.sUint32.nLow);
  DemoPrintTimeStamp(pEventData->m_nTimeStamp);

  switch (pEventData->m_eDeviceType) {
    case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_DSP:
    // Non-logical device related event from the DSPs
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DSP[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("DSP[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_SINGLE_DETECTOR:
#if 0 // temporary
    if (pEventData->m_nSourceId == 1) break;
#endif

    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      // Event from Dtmf Detector
      if (pEventData->m_nCode == OTX_SINGLE_TONE_DETECTOR_EC_TONE) {
        // sqFreq[0] = 440;  // Freq A in Hz
        // sqFreq[1] = 620;  // Freq B in Hz       

        switch (pEventData->m_nParam) {
        case 0:
          DemoLogf("SingleDetector[%d] detected 440Hz Tone (%d) \n", pEventData->m_nSourceId, pEventData->m_nParam);
          break;
        case 1:
          DemoLogf("SingleDetector[%d] detected 620Hz Tone (%d) \n", pEventData->m_nSourceId, pEventData->m_nParam);
          break;
        default:
          DemoLogf("SingleDetector[%d] detected unrecognized tone (%d) \n", pEventData->m_nSourceId, pEventData->m_nParam);
          break;
        }

      } else if (pEventData->m_nCode == OTX_SINGLE_TONE_DETECTOR_EC_PAUSE) {
        DemoLogf("SingleDetector[%d] detected Pause\n", pEventData->m_nSourceId);
      } else if (pEventData->m_nCode == OTX_SINGLE_TONE_DETECTOR_EC_VERBOSE_STATUS) {
        DemoLogf("SingleDetector[%d] Verbose Status 0x%.8X (Code: %d, Value: %d)\n", pEventData->m_nSourceId, pEventData->m_nParam, pEventData->m_nParam & 0x0000FFFF, pEventData->m_nParam >> 16);
      } else {
        // Unexpected event from this device
        DemoLogf("SingleDetector[%d] Unexpected Event: %s \n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("SingleDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}

/*******************************************************************************

  @func

  Uses the SineWaveGenerator device to send a tone (a DTMF tone).

*******************************************************************************/
void DemoSendSineWaveDtmfTone(OTX_INT32 nDialler)
{
  OTX_RESULT result;
  static OTX_INT32 nDtmfDigit = 0;

  // Initialize the SineWave Sender 
  OtxToneSineWaveDataS sqSineWaveData[5];
  
  sqSineWaveData[0].nAttenuation = -9; // -9dB is about 25% of maximum amplitude
  sqSineWaveData[0].nPhase       = 0;  
  
  sqSineWaveData[1].nAttenuation = -9; // -9dB is about 25% of maximum amplitude
  sqSineWaveData[1].nPhase       = 0;  
  
  switch (nDtmfDigit) {
  case 0:
    // DTMF digit 0 
    sqSineWaveData[0].nFrequency   = 941; // in Hz
    sqSineWaveData[1].nFrequency   = 1336; // in Hz       
    break;
  case 1:
    // DTMF digit 1 
    sqSineWaveData[0].nFrequency   = 697; // in Hz
    sqSineWaveData[1].nFrequency   = 1209; // in Hz       
    break;
  case 2:
    // DTMF digit 2 
    sqSineWaveData[0].nFrequency   = 697; // in Hz
    sqSineWaveData[1].nFrequency   = 1336; // in Hz       
    break;
  case 3:
    // DTMF digit 3 
    sqSineWaveData[0].nFrequency   = 697; // in Hz
    sqSineWaveData[1].nFrequency   = 1477; // in Hz       
    break;
  case 4:
    // DTMF digit 4 
    sqSineWaveData[0].nFrequency   = 770; // in Hz
    sqSineWaveData[1].nFrequency   = 1209; // in Hz       
    break;
  case 5:
    // DTMF digit 5 
    sqSineWaveData[0].nFrequency   = 770; // in Hz
    sqSineWaveData[1].nFrequency   = 1336; // in Hz       
    break;
  case 6:
    // DTMF digit 6
    sqSineWaveData[0].nFrequency   = 770; // in Hz
    sqSineWaveData[1].nFrequency   = 1477; // in Hz       
    break;
  case 7:
    // DTMF digit 7 
    sqSineWaveData[0].nFrequency   = 852; // in Hz
    sqSineWaveData[1].nFrequency   = 1209; // in Hz       
    break;
  case 8:
    // DTMF digit 8 
    sqSineWaveData[0].nFrequency   = 852; // in Hz
    sqSineWaveData[1].nFrequency   = 1336; // in Hz       
    break;
  case 9:
    // DTMF digit 9
    sqSineWaveData[0].nFrequency   = 852; // in Hz
    sqSineWaveData[1].nFrequency   = 1477; // in Hz       
    break;
  }
  
  DemoLogf("Initializing SineWave Generator[%d] for DTMF tone %d (%d Hz / %d Hz)\n", nDialler, nDtmfDigit, (OTX_INT32)sqSineWaveData[0].nFrequency, (OTX_INT32)sqSineWaveData[1].nFrequency);
  result = OtxToneSineWaveGeneratorInit(g_hLogicDevs.hSineWaveSender[nDialler], 
    g_demoEnv.law, 
    sqSineWaveData,
    2, // coSineWaveData
    OTX_TONE_GAIN_0DB); DemoCheckRc(result);
  
  
  DemoLogf("Enabling SineWave Generator[%d]... \n", nDialler);
  result = OtxDrvEnable(g_hLogicDevs.hSineWaveSender[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  
  OsaThreadSleep(500);
  
  DemoLogf("Disabling SineWave Generator[%d]... \n", nDialler);
  result = OtxDrvDisable(g_hLogicDevs.hSineWaveSender[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
  
  if (++nDtmfDigit > 9) {
    nDtmfDigit = 0;
  }
  
  

}

/*******************************************************************************

  @func

  Uses the SineWaveGenerator device to send a Single Tone 

*******************************************************************************/
void DemoSendSingleTone(OTX_INT32 nDialler, OTX_INT32 nFrequency)
{
  OTX_RESULT result;
  OTX_INT32 i;

  // Initialize the SineWave Sender 
  OtxToneSineWaveDataS sqSineWaveData[2];
  
  sqSineWaveData[0].nAttenuation = -20; // in dB
  sqSineWaveData[0].nFrequency   = nFrequency; // in Hz
  sqSineWaveData[0].nPhase       = 0;  
  
  DemoLogf("Initializing SineWave Generator[%d] for Test tone (%d Hz)\n", nDialler, (OTX_INT32)sqSineWaveData[0].nFrequency);
  result = OtxToneSineWaveGeneratorInit(g_hLogicDevs.hSineWaveSender[nDialler], 
    g_demoEnv.law, 
    sqSineWaveData,
    1, // coSineWaveData
    OTX_TONE_GAIN_0DB); DemoCheckRc(result);

  // Repeat Testing Tone 3 times
  for (i = 0; i < 3; i++) {
    
    result = OtxDrvEnable(g_hLogicDevs.hSineWaveSender[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);

    DemoLogf("*** Enabled (Sleeping 0.5 seconds)\n");
    OsaThreadSleep(500);
    
    result = OtxDrvDisable(g_hLogicDevs.hSineWaveSender[nDialler],  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    
    DemoLogf("*** Disabled (Sleeping 0.5 seconds)\n");
    OsaThreadSleep(500);
    
  }  
  DemoLogf("SineWave Generator[%d] done\n", nDialler);

}


/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr      = NULL;
  envP->bExitOnError     = OTX_FALSE;
  envP->law              = OTX_DATA_FORMAT_COMPAND8_A_LAW;
  envP->eT1E1Mode        = OTX_T1E1_MODE_E1;
  envP->boardNo          = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType       = OTX_DEVICE_THOR_PCI;
  envP->coDsp            = 2;
  envP->eProgramModule   = OTX_DSP_SPM_OTXSPM2;
  envP->nLinkTimeout     = 3000;


  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'D') {   // DSP Out file (overrides the DSP Program module switch)
          envP->pszDspOutFile = argv[i] + 2;
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == 'T') {
            envP->nBoardType = OTX_DEVICE_THOR_2_PCI_PLUS;
            envP->coLi = 2;
          } else if ( *(argv[i]+2) == 'E') {
            envP->nBoardType = OTX_DEVICE_THOR_2_PCI_EXPRESS;
            envP->coLi = 2;
          } else if ( *(argv[i]+2) == '4') {
            envP->nBoardType = OTX_DEVICE_THOR_4_PCI_PLUS;
            envP->coLi = 2;
          } else if ( *(argv[i]+2) == 'M') {
            envP->nBoardType = OTX_DEVICE_THOR_PCMCIA;
            envP->coLi = 2;
          } else if ( *(argv[i]+2) == 'P') {
            envP->nBoardType = OTX_DEVICE_THOR_8_PCI_PLUS;
            envP->coLi  = 8;
            envP->coDsp = 4;
          } else if ( *(argv[i]+2) == 'W') {
            envP->nBoardType = OTX_DEVICE_THOR_8_PCIE;
            envP->coLi  = 8;
            envP->coDsp = 4;
          } else if ( *(argv[i]+2) == 'S') {
            envP->nBoardType  = OTX_DEVICE_SLEIPNIR_1_PCI;
            envP->coLi = 1;
            envP->coDsp = 1;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {   // E1 
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
          } else if (*(argv[i]+2) == 'T') {   // T1
            envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else if (*(argv[i]+1) == 'P') {   // Dsp Program module
          if (*(argv[i]+2) == '2') {   // OtxSpm2.out
            envP->eProgramModule = OTX_DSP_SPM_OTXSPM2;
          } else if (*(argv[i]+2) == '3') {   // OtxCst3.out
            envP->eProgramModule = OTX_DSP_SPM_OTXCST3;
          } else {
            DemoLogf("Error: Invalid DSP Program Module command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'C') {   // Companding Law
          if (*(argv[i]+2) == 'A') {   // A-law
            envP->law = OTX_DATA_FORMAT_COMPAND8_A_LAW;
          } else if (*(argv[i]+2) == 'U') {   // U-law
            envP->law = OTX_DATA_FORMAT_COMPAND8_U_LAW;
          } else {
            DemoLogf("Error: Invalid companding law command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
				  exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
			  exit(1);
      }
    }
  }

}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /T[T|E|4|M|P|W]: Board Type used for the demo, where\n");
  DemoLogf("                  T = Use Thor-2-PCI-Plus Board \n");
  DemoLogf("                  E = Use Thor-2-PCI-Express Board \n");
  DemoLogf("                  4 = Use Thor-4-PCI-Plus Board \n");
  DemoLogf("                  M = Use Thor-2-PCMCIA Board\n");
  DemoLogf("                  P = Use Thor-8-PCI-Plus Board\n");
  DemoLogf("                  W = Use Thor-8-PCIe Board\n");
  DemoLogf(" /CA          : A-law\n");
  DemoLogf(" /CU          : u-law\n");
  DemoLogf(" /ME          : E1 Li mode\n");
  DemoLogf(" /MT          : T1 Li mode\n");
  DemoLogf(" /P2          : Use OtxSpm2 DSP module (default)\n");
  DemoLogf(" /P3          : Use OtxCst3 DSP module (instead of OtxSpm2 as default)\n");
  DemoLogf(" /Z<str>      : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q           : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(const char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF((char *) formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

}


static void DemoCheckRc(IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {
    OTX_RESULT otxRc;
    char msgStr[256];
    
    // Convert the result code to a more descriptive string
    otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
    if (OTX_FAILURE(otxRc)) {
      strcpy(msgStr, "Cannot translate this result code.");
    }
    DemoLogf("ERROR(%#lX [%d.%d]): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), msgStr);
    
    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  static OTX_HANDLE hCurrentLi = 0;
  static OTX_INT32      nCurrentLi = 0;
  OTX_RESULT result;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" 0-1 Select active Li, this line is then used for subsequent commands\n");
    DemoLogf(" e   Enable board interrupts\n");
    DemoLogf(" E   Disable board interrupts\n");
    DemoLogf(" j   Send one DTMF Tone on TS 5 (from current Li)\n");
		DemoLogf(" l   Query Line Status of the two Lis\n");
    DemoLogf(" r   Send Single Tone 440Hz (3 times) on TS 5 (from current Li)\n");
    DemoLogf(" R   Send Single Tone 470Hz (3 times) on TS 5 (from current Li)\n");
    DemoLogf(" s   Send Single Tone with frequency input from user (3 times) on TS 5 (from current Li)\n");
    break;

  case 'B':
    {
      OTX_INT32 nDsp;
      
      for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
        DemoLogf("Init Bsp for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);
      }
    }
    break;

  case 'j':
    {
      DemoLogf("Sending DTMF Tone on Li#%d TS%d \n", nCurrentLi, 5);

      DemoSendSineWaveDtmfTone(nCurrentLi);

      DemoLogf("Done.\n");
    }
    break;

  case 'r':
    {
      DemoLogf("Sending Test Tone 440Hz (3 times) on Li#%d TS%d \n", nCurrentLi, 5);

      DemoSendSingleTone(nCurrentLi, 440);

      DemoLogf("Done.\n");
    }
    break;

  case 'R':
    {
      DemoLogf("Sending Test Tone 490Hz (3 times) on Li#%d TS%d (should not be detected)\n", nCurrentLi, 5);

      DemoSendSingleTone(nCurrentLi, 490);

      DemoLogf("Done.\n");
    }
    break;

  case 's':
    {
      char strBuf[10];
      static OTX_UINT32 nFreq = 0;

      DemoLogf("Which Frequency: ", nFreq);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nFreq = strtoul(strBuf, NULL, 0);
      }

      DemoLogf("Sending Test Tone %dHz (3 times) on Li#%d TS%d \n", nFreq, nCurrentLi, 5);

      DemoSendSingleTone(nCurrentLi, nFreq);

      DemoLogf("Done.\n");
    }
    break;

  case 'e':
    DemoLogf("Enabling board interrupts\n");
    OtxDrvEnable(g_hPhysDevs.hThorController0, DEMO_TASK_REF, OTX_TIME_INFINITY);
    break;

  case 'E':
    DemoLogf("Disabling board interrupts\n");
    OtxDrvDisable(g_hPhysDevs.hThorController0, DEMO_TASK_REF, OTX_TIME_INFINITY);
    break;

  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 i;

      for (i = 0; i < g_demoEnv.coLi; i++) {
        DemoLogf("Line status of Li #%d is ", i);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[i], &eLayer1Status); DemoCheckRc(result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }

    }
    break;

  case '0':
    // Select Li0/Dsp0 as current
    nCurrentLi = 0;
    hCurrentLi = g_hPhysDevs.hLi[nCurrentLi];
    DemoLogf("LI#%d Selected ....\n", nCurrentLi);
    break;

  case '1':
    // Select Li1/Dsp1 as current
    nCurrentLi = 1;
    hCurrentLi = g_hPhysDevs.hLi[nCurrentLi];
    DemoLogf("LI#%d Selected ....\n", nCurrentLi);
    break;

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}


/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  DemoLogf("SingleTone, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("Single tone generate/monitor example program.\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(result);
    break;
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Thor-2-PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();

  // Enable all physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling physical devices...\n");
  DemoEnablePhysicalDevices();

  // Set Li0 as the clock source
  if (g_hPhysDevs.hVidarAsmController0) {
    // Thor-2-PCI, Thor-8-PCI
    result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_LOCAL_0); DemoCheckRc(result);
#if 1  // Temporary
    {
      OTX_INT32 nDsp;
      for (nDsp = 0; nDsp < g_demoEnv.coDsp; nDsp++) {
        DemoLogf("Init Bsp for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);
      }
    }
#endif
  } else {
    // Thor-2-PCMCIA
    result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_NA, OTX_CLOCK_SOURCE_LOCAL_0); DemoCheckRc(result);
  }

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("Connecting devices...\n");
  DemoConnectDevices();

  // Initialize the Connected Logical Devices
  DemoLogf("Initializing logical devices...\n");
  DemoInitializeLogicalDevices();

  // Enable all logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling logical devices...\n");
  DemoEnableLogicalDevices();

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
	OsaThreadStart(DemoIdleFunction);

  OsaThreadSleep(300); // To let DemoIdleFunction start up
  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OtxDrvReleaseLib();
	OsaConsoleExit();

  return 0;
}

